"""DNS Authenticator for Active24 DNS."""
import logging

import requests

import zope.interface

import signal

import dns
import dns.message
import dns.name
import dns.query
import dns.rdatatype
import dns.resolver

from time import sleep

from certbot import errors
from certbot import interfaces
from certbot.plugins import dns_common

logger = logging.getLogger(__name__)


@zope.interface.implementer(interfaces.IAuthenticator)
@zope.interface.provider(interfaces.IPluginFactory)
class Authenticator(dns_common.DNSAuthenticator):
    """DNS Authenticator for Active24 DNS

    This Authenticator uses the Active24 DNS API to fulfill a dns-01 challenge.
    """

    description = 'Obtain certificates using a DNS TXT record (if you are using Active24 for DNS).'

    def __init__(self, *args, **kwargs):
        super(Authenticator, self).__init__(*args, **kwargs)
        self.credentials = None

    @classmethod
    def add_parser_arguments(cls, add):  # pylint: disable=arguments-differ
        super(Authenticator, cls).add_parser_arguments(add, default_propagation_seconds=0)
        add('credentials', help='Path to Active24 credentials INI file', default='/etc/letsencrypt/active24.ini')

    def more_info(self):  # pylint: disable=missing-docstring,no-self-use
        return 'This plugin configures a DNS TXT record to respond to a dns-01 challenge using ' + \
               'the Active24 API.'

    def _setup_credentials(self):
        self.credentials = self._configure_credentials(
            'credentials',
            'path to Active24 credentials INI file',
            {
                'token': 'API token for the Active24 account.',
            }
        )

    def perform(self, achalls):
        responses = super(Authenticator, self).perform(achalls)

        if self.conf('propagation-seconds') <= 0:
            _wait_for_propagation(achalls)

        return responses

    def _perform(self, domain, validation_name, validation):
        self._get_active24_client().add_txt_record(validation_name, validation)

    def _cleanup(self, domain, validation_name, validation):
        self._get_active24_client().del_txt_record(validation_name, validation)

    def _get_active24_client(self):
        return _Active24Client(self.credentials.conf('token'))


def _wait_for_propagation(challenges):
    queue = [(ch.validation_domain_name(ch.domain), ch.validation(ch.account_key)) for ch in challenges]
    queue = [(o[0], o[1], dns.message.make_query(dns.name.from_text(o[0]), dns.rdatatype.TXT)) for o in queue]
    queue = [(ns, o[1], o[2]) for o in queue for ns in _get_nameservers(o[0])]

    logger.debug('Waiting for propagation to authoritative servers')
    i = 0

    def break_loop(*_):
        logger.debug('Interrupted by user signal')
        queue.clear()

    orig = signal.signal(signal.SIGUSR1, break_loop)

    while len(queue) > 0:
        queue = [(ns, content, query) for (ns, content, query) in queue if not _check_nameserver(ns, query, content)]
        sleep(1)
        i += 1

        if (i % 30) == 0:
            logger.debug('Remaining records to check: %d' % len(queue))

    signal.signal(signal.SIGUSR1, orig)


def _check_nameserver(ns, query, content):
    result = dns.query.udp(query, ns)

    try:
        answers = [a for r in result.answer for a in r]

        for a in answers:
            value = ''.join([s.decode('utf-8') for s in a.strings])

            if value == content:
                return True
    except (KeyError, IndexError):
        return False

    return False


def _get_nameservers(domain):
    resolver = dns.resolver.get_default_resolver()

    while True:
        try:
            rrset = resolver.query(domain, 'NS')
        except (dns.resolver.NoAnswer, dns.resolver.NXDOMAIN):
            try:
                domain = domain[domain.index('.') + 1:]
                continue
            except ValueError:
                return []

        if len(rrset) == 0:
            return []

        nameservers = [resolver.query(rr.target.to_text(), raise_on_no_answer=False) for rr in rrset]
        return [ns[0].to_text() for ns in nameservers if len(ns) > 0]


class _Active24Client(object):
    """
    Encapsulates all communication with the Active24
    """

    def __init__(self, token):
        self.token = token if isinstance(token, str) else ','.join(token)
        self.test = False

    def add_txt_record(self, record_name, record_content):
        """
        Add a TXT record using the supplied information.
        :param str record_name: The record name (typically beginning with '_acme-challenge.').
        :param str record_content: The record content (typically the challenge validation).
        :raises certbot.errors.PluginError: if an error occurs communicating with the Active24 API
        """

        domain, record = self._parse_domain(record_name)
        logger.debug('Attempting to add record: %s' % record)

        try:
            response = self._send_request('POST', '/dns/%s/txt/v1' % domain, {
                'name': record,
                'text': record_content,
                'ttl': 300,
            })
        except requests.exceptions.RequestException as e:
            logger.error('Encountered error adding TXT record: %s', e)

            if e.response is not None:
                logger.error(e.response.text)

            raise errors.PluginError('Error communicating with the Active24 API: %s' % e)

        if response.status_code != 204:
            logger.error('Encountered error adding TXT record: %s' % response)
            raise errors.PluginError('Error communicating with the Active24 API: %s' % response)

        logger.debug('Successfully added TXT record')

    def del_txt_record(self, record_name, record_content):
        """
        Delete a TXT record using the supplied information.
        Note that both the record's name and content are used to ensure that similar records
        created concurrently (e.g., due to concurrent invocations of this plugin) are not deleted.
        Failures are logged, but not raised.
        :param str record_name: The record name (typically beginning with '_acme-challenge.').
        :param str record_content: The record content (typically the challenge validation).
        """

        domain, record = self._parse_domain(record_name)
        logger.debug('Attempting to delete record: %s' % record)
        dns_record = self._find_record(domain, record, record_content)

        if dns_record is not None:
            try:
                response = self._send_request('DELETE', '/dns/%s/%s/v1' % (domain, dns_record['hashId']))
            except requests.exceptions.RequestException as e:
                logger.warning('Encountered error deleting TXT record: %s' % e)
                return
        else:
            logger.debug('Record doesn\'t appear to exist, aborting')
            return

        if response.status_code != 204:
            logger.warning('Encountered error deleting TXT record: %s' % response)
            return

        logger.debug('Successfully deleted TXT record.')

    def _find_record(self, domain_name, record_name, record_content):
        response = self._send_request('GET', '/dns/%s/records/v1' % domain_name)
        records = response.json()

        for record in records:
            if record['type'] == 'TXT' and record['name'] == record_name and record['text'] == record_content:
                return record

        return None

    def _parse_domain(self, domain):
        """
        Parses full domain into base domain name and record name
        :param str domain: Domain name
        :returns: Top-level
        :rtype: tuple
        """
        pieces = domain.split('.')
        return '.'.join(pieces[-2:]), '.'.join(pieces[:-2])

    def _send_request(self, method, endpoint, payload=None):
        base_url = 'https://sandboxapi.active24.com' if self.test else 'https://api.active24.com'

        response = requests.request(
            method,
            base_url + endpoint,
            json=payload,
            headers={'Authorization': 'Bearer %s' % self.token}
        )

        response.raise_for_status()
        return response



from setuptools import setup
from setuptools import find_packages

from certbot_dns_active24 import __version__

install_requires = [
    'acme>=0.21.1',
    'certbot>=0.21.1',
    'dnspython>=1.16.0',
    'requests>=2.9.1',
    'mock',
    'setuptools',
    'zope.interface',
]

data_files = [
    ('/etc/letsencrypt', ['active24.ini'])
]

with open('README.md') as f:
    long_description = f.read()

setup(
    name='certbot-dns-active24',
    version=__version__,
    description="Active24 DNS authenticator plugin for Certbot",
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords='certbot letsencrypt dns-01 plugin',
    url='https://github.com/jahudka/certbot-dns-active24',
    download_url='https://github.com/jahudka/certbot-dns-active24/archive/v' + __version__ + '.tar.gz',
    author="Dan Kadera",
    author_email='me@subsonic.cz',
    license='MIT',
    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Plugins',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Security',
        'Topic :: System :: Installation/Setup',
        'Topic :: System :: Networking',
        'Topic :: System :: Systems Administration',
        'Topic :: Utilities',
    ],
    install_requires=install_requires,
    data_files=data_files,
    packages=find_packages(),
    include_package_data=True,
    entry_points={
        'certbot.plugins': [
            'dns-active24 = certbot_dns_active24.dns_active24:Authenticator',
        ],
    },
    test_suite='certbot_dns_active24',
)



"""Tests for certbot_dns_active24.dns."""

import os
import unittest

import mock
import requests

from certbot import errors
from certbot.plugins import dns_test_common
from certbot.plugins.dns_test_common import DOMAIN
from certbot.tests import util as test_util

TOKEN = '123456qwerty-ok'

HTTP_ERROR = requests.exceptions.RequestException()


class AuthenticatorTest(test_util.TempDirTestCase, dns_test_common.BaseAuthenticatorTest):

    def setUp(self):
        from certbot_dns_active24.dns_active24 import Authenticator

        super(AuthenticatorTest, self).setUp()

        path = os.path.join(self.tempdir, 'file.ini')
        dns_test_common.write({"active24_token": TOKEN}, path)

        self.config = mock.MagicMock(active24_credentials=path, active24_propagation_seconds=0)  # don't wait during tests

        self.auth = Authenticator(self.config, "active24")

        self.mock_client = mock.MagicMock()
        # _get_active24_client | pylint: disable=protected-access
        self.auth._get_active24_client = mock.MagicMock(return_value=self.mock_client)

    def test_perform(self):
        self.auth.perform([self.achall])

        expected = [mock.call.add_txt_record('_acme-challenge.' + DOMAIN, mock.ANY)]
        self.assertEqual(expected, self.mock_client.mock_calls)

    def test_cleanup(self):
        # _attempt_cleanup | pylint: disable=protected-access
        self.auth._attempt_cleanup = True
        self.auth.cleanup([self.achall])

        expected = [mock.call.del_txt_record('_acme-challenge.' + DOMAIN, mock.ANY)]
        self.assertEqual(expected, self.mock_client.mock_calls)


class Active24ClientTest(unittest.TestCase):

    record_prefix = "_acme-challenge"
    record_name = record_prefix + "." + DOMAIN
    record_content = "test"

    def setUp(self):
        from certbot_dns_active24.dns_active24 import _Active24Client

        self.client = _Active24Client(TOKEN)

        # _send_request | pylint: disable=protected-access
        self.client._send_request = mock.Mock()

    def test_add_txt_record(self):
        self.client._send_request.side_effect = [self._create_response(204)]
        self.client.add_txt_record(self.record_name, self.record_content)
        self.client._send_request.assert_called_with('POST', '/dns/{0}/txt/v1'.format(DOMAIN), {
            'name': self.record_prefix,
            'text': self.record_content,
            'ttl': 300,
        })

    def test_add_txt_record_subdomain(self):
        self.client._send_request.side_effect = [self._create_response(204)]
        self.client.add_txt_record(self.record_prefix + '.subdomain.' + DOMAIN, self.record_content)
        self.client._send_request.assert_called_with('POST', '/dns/{0}/txt/v1'.format(DOMAIN), {
            'name': self.record_prefix + '.subdomain',
            'text': self.record_content,
            'ttl': 300,
        })

    def test_add_txt_record_error_send_request(self):
        self.client._send_request.side_effect = HTTP_ERROR
        self.assertRaises(errors.PluginError, self.client.add_txt_record, self.record_name, self.record_content)

    def test_del_txt_record(self):
        self.client._send_request.side_effect = [
            self._create_response(200, [{'name': self.record_prefix, 'hashId': 'asdfqwert'}]),
            self._create_response(204)
        ]
        self.client.del_txt_record(self.record_name, self.record_content)

    def test_del_txt_record_subdomain(self):
        self.client._send_request.side_effect = [
            self._create_response(200, [{'name': self.record_prefix + '.subdomain', 'hashId': 'asdfqwert'}]),
            self._create_response(204)
        ]
        self.client.del_txt_record(self.record_prefix + '.subdomain.' + DOMAIN, self.record_content)

    def _create_response(self, status=200, payload=None):
        def rfs():
            if status >= 400:
                raise requests.exceptions.HTTPError()

        res = mock.Mock()
        res.status_code = status
        res.json = lambda: payload
        res.raise_for_status = rfs
        return res


if __name__ == "__main__":
    unittest.main()  # pragma: no cover
